/* mini-at/src/mini-at/main.c
 *
 *  (c)2006, Laurence Withers, <l@lwithers.me.uk>.
 *  Released under the GNU GPLv2. See file COPYING or
 *  http://www.gnu.org/copyleft/gpl.html for details.
*/



void msg_add(char* buf, int* pos, const char* src, int amt)
{
    if((*pos + amt) > AT_MSG_SIZE) {
        fprintf(stderr, "Message size too long: try reducing number of commandline arguments.\n");
        exit(1);
    }

    memcpy(buf + *pos, src, amt);
    *pos += amt;
}



void version(void)
{
    printf("mini-at " VERSION "\n"
        "(c)2006, Laurence Withers.\n");
}



void usage(void)
{
    printf("Usage:\n\n"
        "    mini-at [options] time cmd [args ...]\n\n"
        "Options:\n"
        "-h, --help         Display this screen.\n"
        "-V, --version      Display version.\n"
        "-f, --file <x>     Specify filename to read stdin from, - to copy stdin.\n"
        "-0, --argv0 <x>    Change argv[0] from `cmd' to `x'.\n"
        "-s, --socket <x>   Name of mini-atd's socket (default: ~/.mini-atd).\n"
        "\n"
        "'time' is an ISO8601 time. If you specify an incomplete time, e.g. '2010', we simply\n"
        "take the earliest time it could represent (i.e. 2010-01-01T00:00:00).\n"
        "\n");
    version();
}



struct option options[] = {
    { "help", no_argument, 0, 'h' },
    { "version", no_argument, 0, 'V' },
    { "file", required_argument, 0, 'f' },
    { "argv0", required_argument, 0, '0' },
    { 0, 0, 0, 0 }
};



int main(int argc, char* argv[])
{
    const char* stdin_file = 0, * argv0 = 0, * socket = 0;
    struct at_msg a;
    char msg[AT_MSG_SIZE], sockbuf[sizeof(struct sockaddr_un) - sizeof(int)];
    int size, i, x;
    time_t t;

    while(1) {
        switch(getopt_long(argc, argv, "hVf:0:s:", options, 0)) {
        case -1:
            goto done;

        case 'h':
            usage();
            return 0;

        case 'V':
            version();
            return 0;

        case 'f':
            if(stdin_file) {
                fprintf(stderr, "Input file can only be specified once.\n");
                return 1;
            }
            stdin_file = optarg;
            break;

        case '0':
            if(argv0) {
                fprintf(stderr, "argv[0] can only be specified once.\n");
                return 1;
            }
            argv0 = optarg;
            break;

        case 's':
            if(socket) {
                fprintf(stderr, "Socket can only be specified once.\n");
                return 1;
            }
            socket = optarg;
            break;
        }
    }

done:
    argv += optind;
    argc -= optind;
    if(argc < 2) {
        usage();
        return 1;
    }

    // parse time
    t = parse_time(argv[0]);
    if(t == (time_t)-1) {
        fprintf(stderr, "Not a valid ISO8601 date: ``%s''.\n"
            "See e.g. http://en.wikipedia.org/wiki/Iso8601 .\n",
            argv[0]);
        return 1;
    }
    ++argv;
    --argc;

    // set up message header
    memset(&a, 0, sizeof(a));
    a.magic = AT_MAGIC;
    a.version = (VERMAJOR << 20) | (VERMINOR << 10) | VERMICRO;
    a.when = t;
    a.cmd_len = strlen(argv[0]);
    if(stdin_file) {
        stdin_file = do_stdin(stdin_file);
        if(!stdin_file) return 1;
        a.in_fn_len = strlen(stdin_file);
    }
    a.num_args = argc;

    // set up message
    size = 0;
    msg_add(msg, &size, (char*)&a, sizeof(a));
    msg_add(msg, &size, argv[0], a.cmd_len);
    if(stdin_file) msg_add(msg, &size, stdin_file, a.in_fn_len);

    if(!argv0) argv0 = argv[0];
    x = strlen(argv0);
    msg_add(msg, &size, (char*)&x, sizeof(x));
    msg_add(msg, &size, argv0, x);

    for(i = 1; i < argc; ++i) {
        x = strlen(argv[i]);
        msg_add(msg, &size, (char*)&x, sizeof(x));
        msg_add(msg, &size, argv[i], x);
    }

    a.magic = ~a.magic;
    msg_add(msg, &size, (char*)&a.magic, sizeof(a.magic));

    // send message
    if(!socket) {
        socket = getenv("HOME");
        if(!socket) {
            fprintf(stderr, "$HOME not set, cannot find socket\n");
            return 1;
        }
        snprintf(sockbuf, sizeof(sockbuf) - 1, "%s/.mini-atd", socket);
        sockbuf[sizeof(sockbuf) - 1] = 0;
        socket = sockbuf;
    }
    return send_message(socket, msg, size);
}



/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
