/* mini-at/src/mini-atd/socket.c
 *
 *  (c)2006, Laurence Withers, <l@lwithers.me.uk>.
 *  Released under the GNU GPLv2. See file COPYING or
 *  http://www.gnu.org/copyleft/gpl.html for details.
*/



/* process_message()
 *
 *  This is called whenever we receive a single message. It should decode the message and add an
 *  entry to the time database `tdb'.
 */
void process_message(const char* buf, int size, struct time_db* tdb)
{
    struct at_msg a;
    struct time_db_entry t;
    struct tm* tm;
    int i, j, realsize;

    realsize = size;

    memset(&t, 0, sizeof(t));

    // sanity checking
    if(size < (int)sizeof(struct at_msg)) goto bad;
    memcpy(&a, buf, sizeof(a));

    if(a.magic != AT_MAGIC) {
        LOG(LOG_WARNING, "bad message received through socket (magic was 0x%X, should be 0x%X).",
            a.magic, AT_MAGIC);
        return;
    }

    if(a.version != ((VERMAJOR << 20) | (VERMINOR << 10) | VERMICRO)) {
        LOG(LOG_WARNING, "version mismatch (got %d.%d.%d, expected " VERSION ").",
            a.version >> 20, (a.version >> 10) & 0x3FF, a.version & 0x3FF);
        return;
    }

    t.when = a.when;
    t.num_args = a.num_args;
    buf += sizeof(struct at_msg);
    size -= sizeof(struct at_msg);

    // sanity check
    if(a.cmd_len < 1 || a.in_fn_len < 0 || a.num_args < 1
        || size < a.cmd_len + a.in_fn_len) goto bad;

    // read in the command path
    t.path = safe_malloc(a.cmd_len + 1);
    memcpy(t.path, buf, a.cmd_len);
    buf += a.cmd_len;
    size -= a.cmd_len;

    // read in (optional) input filename
    if(a.in_fn_len) {
        t.in = safe_malloc(a.in_fn_len + 1);
        memcpy(t.in, buf, a.in_fn_len);
        buf += a.in_fn_len;
        size -= a.in_fn_len;
    }

    // now read in arguments
    t.args = safe_malloc(sizeof(char*) * (t.num_args + 1)); // extra +1 for NULL for execve
    for(i = 0; i < t.num_args; ++i) {
        if(size < (int)sizeof(int)) goto bad;
        memcpy(&j, buf, sizeof(int));
        if(j < 0) goto bad;
        buf += sizeof(int);
        size -= sizeof(int);

        if(size < j) goto bad;
        t.args[i] = safe_malloc(j + 1);
        if(j) {
            memcpy(t.args[i], buf, j);
            buf += j;
            size -= j;
        }
    }

    // check for end
    if(size != sizeof(int)) goto bad;
    memcpy(&j, buf, sizeof(int));
    if(j != ~AT_MAGIC) goto bad;

    // done -- insert our new entry
    tm = gmtime(&t.when);
    LOG(LOG_INFO, "Command ``%s'' scheduled for %04d-%02d-%02dT%02d:%02d:%02d",
        t.path, tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
    add_time_db_entry(tdb, &t);
    return;

bad:
    LOG(LOG_WARNING, "bad message of %d bytes received through socket.", realsize);
}



/* process_socket()
 *
 *  Checks the socket for input; if some is available, processes it.
 */
int process_socket(int fd, struct time_db* tdb)
{
    ssize_t rd;
    char buf[AT_MSG_SIZE];

    while(1) {
        rd = read(fd, buf, sizeof(buf));
        if(rd == -1) {
            if(errno == EINTR) continue;
            if(errno == EAGAIN) return 0;

            LOG(LOG_CRIT, "read() on socket fd %d failed - %s (%d).",
                fd, strerror(errno), errno);
            return -1;
        }

        process_message(buf, rd, tdb);
    }
}



/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
