/* mini-at/src/mini-atd/time.c
 *
 *  (c)2006, Laurence Withers, <l@lwithers.me.uk>.
 *  Released under the GNU GPLv2. See file COPYING or
 *  http://www.gnu.org/copyleft/gpl.html for details.
*/



/* struct time_db_entry
 *
 *  A structure which holds a single entry for a command. It has the date at which the command will
 *  be run, the command, its arguments, and potentially a filename from which stdin will be 
 *  redirected. It is part of a linked list.
 */
struct time_db_entry {
    time_t when;            // when do we run it?
    char* path;             // path to program
    int num_args;           // number of arguments
    char** args;            // list of arguments
    char* in;               // possible input filename

    struct time_db_entry* next;
};



/* struct time_db
 *
 *  Holds a list of commands. The list must be sorted. Also holds a variable array of PIDs which we 
 *  are waiting on for exit status.
 */
struct time_db {
    struct time_db_entry* head;

    pid_t* pids, * pids2;
    int num_pids, sz_pids;
};



/* free_time_db_entry()
 *
 *  Frees a time database entry (frees memory associated with it and unlinks any temporary files
 *  associated with it).
 */
void free_time_db_entry(struct time_db_entry* t)
{
    int i;

    free(t->path);
    for(i = 0; i < t->num_args; ++i) free(t->args[i]);
    free(t->args);
    if(t->in) unlink(t->in);
    free(t->in);

    free(t);
}



/* process_time_db_entry()
 *
 *  Runs a time database entry.
 */
pid_t process_time_db_entry(const struct time_db_entry* t)
{
    pid_t pid;

    LOG(LOG_INFO, "Running program ``%s'' (%d args).", t->path, t->num_args);

    switch( (pid = fork()) ) {
    case -1:
        LOG(LOG_ALERT, "fork() failed - %s (%d) - couldn't run command.", strerror(errno), errno);
        return -1;

    case 0:
        // child
        break;

    default:
        // parent -- return PID
        return pid;
    }

    // if we have an input file, we should redirect stdin from there
    if(t->in) {
        close(0);
        if(open(t->in, O_RDONLY) == -1) {
            LOG(LOG_ERR, "Couldn't redirect stdin from ``%s'' for command ``%s'' - %s (%d).",
                t->in, t->path, strerror(errno), errno);
            _exit(1); // this is the child we are exiting, not the parent
        }
    }

    // now run the new process, inheriting the current environment
    execve(t->path, t->args, environ);
    LOG(LOG_ERR, "Couldn't run command ``%s'' - %s (%d).",
        t->path, strerror(errno), errno);
    _exit(1); // child, not parent
}



/* process_time_db()
 *
 *  Processes the time database structure `tdb', running all commands for which we have entries, and
 *  removing them from the list. Also wait()s on any PIDs we have spawned.
 */
void process_time_db(struct time_db* tdb)
{
    struct time_db_entry* iter, * next;
    time_t now;
    int i, j, status;
    pid_t pid, * pids;

    // clean up children
    for(j = 0, i = 0; i < tdb->num_pids; ++i) {
        switch(waitpid(tdb->pids[i], &status, WNOHANG)) {
        case 0: // nothing exited
            // copy to new list
            tdb->pids2[j++] = tdb->pids[i];
            break;

        case -1: // error
            LOG(LOG_WARNING, "waitpid(%d) returned error - %s (%d).",
                tdb->pids[i], strerror(errno), errno);
            break;

        default: // exited
            if(WIFSIGNALED(status)) {
                LOG(LOG_INFO, "process %d killed by signal %d.", 
                    tdb->pids[i], WTERMSIG(status));
            } else if(WIFEXITED(status)) {
                LOG(LOG_INFO, "process %d exited with status %d.", 
                    tdb->pids[i], WEXITSTATUS(status));
            } else {
                LOG(LOG_INFO, "process %d exited via unknown method %x.",
                    tdb->pids[i], status);
            }
            break; // we no longer track this PID
        }
    }
    // swap tables
    pids = tdb->pids;
    tdb->pids = tdb->pids2;
    tdb->pids2 = pids;
    tdb->num_pids = j;

    // process the list of commands (sorted)
    time(&now);

    iter = tdb->head;
    while(iter && now >= iter->when) {
        // run this command
        pid = process_time_db_entry(iter);
        if(pid != -1) {
            // add to children list
            if(tdb->num_pids == tdb->sz_pids) {
                // resize array
                if(tdb->sz_pids) tdb->sz_pids <<= 1;
                else tdb->sz_pids = 4;

                tdb->pids = safe_realloc(tdb->pids, tdb->sz_pids * sizeof(pid_t));
                tdb->pids2 = safe_realloc(tdb->pids2, tdb->sz_pids * sizeof(pid_t));
            }

            tdb->pids[tdb->num_pids++] = pid;
        }

        // free up from linked list
        next = iter->next;
        free_time_db_entry(iter);
        tdb->head = next;
        iter = next;
    }
}



/* free_time_db()
 *
 *  Frees the time database and any entries associated with it. This will clean up all temporary
 *  files used for stdin redirection.
 */
void free_time_db(struct time_db* tdb)
{
    struct time_db_entry* iter, * next;

    for(iter = tdb->head; iter; iter = next) {
        next = iter->next;
        free_time_db_entry(iter);
    }

    free(tdb->pids);
    free(tdb->pids2);
    free(tdb);
}



/* add_time_db_entry()
 *
 *  Adds an entry `t' into the time database `tdb'. Bitwise copies `t'. Does a sorted list insert.
 */
void add_time_db_entry(struct time_db* tdb, const struct time_db_entry* t)
{
    struct time_db_entry* iter, * prev, * new;

    new = safe_malloc(sizeof(struct time_db_entry));
    memcpy(new, t, sizeof(struct time_db_entry));

    prev = 0;
    iter = tdb->head;
    while(iter) {
        if(t->when <= iter->when) break;
        prev = iter;
        iter = iter->next;
    }
    new->next = iter;

    if(prev) prev->next = new;
    else tdb->head = new;
}



struct time_db* _debug_time_db = 0;
void dump_time_db(struct time_db* tdb)
{
    int i;
    struct tm tm;
    struct time_db_entry* iter;

    for(i = 0; i < tdb->num_pids; ++i) LOG(LOG_DEBUG, "waiting on PID %lu", (long)tdb->pids[i]);
    for(iter = tdb->head; iter; iter = iter->next) {
        gmtime_r(&iter->when, &tm);
        LOG(LOG_DEBUG, "time database has entry %s for %04d%02d%02dT%02d%02d%02d",
            iter->path, tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
    }
}

void debug_handler(int sig)
{
    signal(sig, debug_handler);
    if(_debug_time_db) dump_time_db(_debug_time_db);
    else LOG(LOG_DEBUG, "no time database");
}



/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
